@@ -1,11 +1,12 @@
 """ flywheel """
-__version__ = '0.4.0'
-
 from dynamo3 import (CheckFailed, ConditionalCheckFailedException, Binary,
                      STRING, NUMBER, BINARY, STRING_SET, NUMBER_SET,
-                     BINARY_SET)
+                     BINARY_SET, Limit)
 
-from .fields import Field, Composite, GlobalIndex, TypeDefinition, set_
+from .fields import Field, Composite, GlobalIndex
+from .fields.types import set_, TypeDefinition
 from .models import Model
 from .engine import Engine
 from .query import DuplicateEntityException, EntityNotFoundException
+
+__version__ = '0.5.4'
