@@ -1,16 +1,22 @@
 """ Query engine """
 import itertools
-from collections import defaultdict
 
+import logging
 import six
-from dynamo3 import DynamoDBConnection, CheckFailed, ItemUpdate, ALL_NEW
-from six.moves import zip as izip  # pylint: disable=F0401
+from collections import defaultdict
+from dynamo3 import (DynamoDBConnection, CheckFailed, ItemUpdate, ALL_NEW,
+                     UPDATED_NEW)
 
 from .fields import Field
 from .models import Model, SetDelta
 from .query import Query, Scan
 
 
+LOG = logging.getLogger(__name__)
+
+NO_ARG = object()
+
+
 class Engine(object):
 
     """
@@ -136,12 +142,16 @@ class Engine(object):
 
     def connect_to_region(self, region, **kwargs):
         """ Connect to an AWS region """
-        self.dynamo = DynamoDBConnection.connect_to_region(region, **kwargs)
+        self.dynamo = DynamoDBConnection.connect(region, **kwargs)
 
     def connect_to_host(self, **kwargs):
         """ Connect to a specific host """
         self.dynamo = DynamoDBConnection.connect_to_host(**kwargs)
 
+    def connect(self, *args, **kwargs):
+        """ Connect to a specific host """
+        self.dynamo = DynamoDBConnection.connect(*args, **kwargs)
+
     def register(self, *models):
         """
         Register one or more models with the engine
@@ -205,6 +215,53 @@ class Engine(object):
                 changed.append(result)
         return changed
 
+    def update_schema(self, test=False, throughput=None):
+        """
+        Updates the DynamoDB table global indexes required by the registered models
+
+        Parameters
+        ----------
+        test : bool, optional
+            If True, perform a dry run (default False)
+        throughput : dict, optional
+            If provided, override the throughputs of the Models when creating
+            them. Details below.
+
+        Returns
+        -------
+        names : list
+            List of table names that were updated
+
+        Examples
+        --------
+        The ``throughput`` argument is a mapping of table names to their
+        throughputs. The throughput is a dict with a 'read' and 'write' value.
+        It may also include the names of global indexes that map to their own
+        dicts with a 'read' and 'write' value.
+
+        .. code-block:: python
+
+            engine.create_schema(throughput={
+                'table1': {
+                    'gindex-1': {
+                        'read': 6,
+                        'write': 3,
+                    }
+                }
+            })
+
+        """
+        throughput = throughput or {}
+        changed = []
+        for model in six.itervalues(self.models):
+            result = model.meta_.update_dynamo_schema(
+                self.dynamo, test=test, wait=True,
+                throughput=throughput.get(model.meta_.ddb_tablename()),
+                namespace=self.namespace)
+            if result:
+                changed.append(result)
+        return changed
+
     def delete_schema(self, test=False):
         """
         Drop the DynamoDB tables for all registered models
@@ -226,7 +283,8 @@ class Engine(object):
             result = model.meta_.delete_dynamo_schema(self.dynamo, tablenames,
                                                       test=test, wait=True,
                                                       namespace=self.namespace)
-            changed.append(result)
+            if result:
+                changed.append(result)
         return changed
 
     def get_schema(self):
@@ -402,7 +460,7 @@ class Engine(object):
                     expected = item.construct_ddb_expects_()
                     count += 1
                     self.dynamo.delete_item(tablename, item.pk_dict_,
-                                            expected=expected)
+                                            **expected)
             else:
                 with self.dynamo.batch_write(tablename) as batch:
                     for item in items:
@@ -483,19 +541,30 @@ class Engine(object):
             return
 
         tables = defaultdict(list)
+        model_map = defaultdict(dict)
         for item in items:
-            tables[item.meta_.ddb_tablename(self.namespace)].append(item)
+            tablename = item.meta_.ddb_tablename(self.namespace)
+            tables[tablename].append(item)
+            model_map[tablename][item.pk_tuple_] = item
 
         for tablename, items in six.iteritems(tables):
             keys = [item.pk_dict_ for item in items]
             results = self.dynamo.batch_get(tablename, keys,
                                             consistent=consistent)
-            for item, data in izip(items, results):
+            meta = items[0].meta_
+            for result in results:
+                pkey = meta.pk_tuple(None, result)
+                item = model_map[tablename].get(pkey)
+                if item is None:
+                    LOG.error("Refresh error: Cannot match primary key %r to "
+                              "a model", pkey)
+                    continue
                 with item.loading_(self):
-                    for key, val in data.items():
+                    for key, val in six.iteritems(result):
                         item.set_ddb_val_(key, val)
 
-    def sync(self, items, raise_on_conflict=None, consistent=False, constraints=None):
+    def sync(self, items, raise_on_conflict=None, consistent=False,
+             constraints=None, no_read=False):
         """
         Sync model changes back to database
 
@@ -517,11 +586,14 @@ class Engine(object):
             List of more complex constraints that must pass for the update to
             complete. Must be used with raise_on_conflict=True. Format is the
             same as query filters (e.g. Model.fieldname > 5)
+        no_read : bool, optional
+            If True, don't perform a GET on models with no changes. (default False)
 
         Raises
         ------
         exc : :class:`dynamo3.CheckFailed`
-            If raise_on_conflict=True and the model changed underneath us
+            If raise_on_conflict=True and the data in dynamo fails the
+            contraint checks.
 
         """
         if raise_on_conflict is None:
@@ -533,18 +605,14 @@ class Engine(object):
         refresh_models = []
         for item in items:
             # Look for any mutable fields (e.g. sets) that have changed
-            for name in item.keys_():
-                field = item.meta_.fields.get(name)
-                if field is None:
-                    value = item.get_(name)
-                    if Field.is_overflow_mutable(value):
-                        if value != item.cached_(name):
-                            item.__dirty__.add(name)
-                elif (field.is_mutable and name not in item.__dirty__ and
-                        name not in item.__incrs__):
+            for name, field in six.iteritems(item.meta_.fields):
+                if name in item.__dirty__ or name in item.__incrs__:
+                    continue
+                if field.is_mutable:
                     cached_var = item.cached_(name)
                     if field.resolve(item) != cached_var:
-                        item.__dirty__.add(name)
+                        for related in item.meta_.related_fields[name]:
+                            item.__dirty__.add(related)
 
             if not item.__dirty__ and not item.__incrs__:
                 refresh_models.append(item)
@@ -552,20 +620,9 @@ class Engine(object):
             fields = item.__dirty__
             item.pre_save_(self)
 
-            # If the model has changed any field that is part of a composite
-            # field, FORCE the sync to raise on conflict. This prevents the
-            # composite key from potentially getting into an inconsistent state
-            _raise_on_conflict = raise_on_conflict
-            for name in itertools.chain(item.__incrs__, fields):
-                for related_name in item.meta_.related_fields.get(name, []):
-                    field = item.meta_.fields[related_name]
-                    if field.composite:
-                        _raise_on_conflict = True
-                        break
-
             keywords = {}
             constrained_fields = set()
-            if _raise_on_conflict and constraints is not None:
+            if raise_on_conflict and constraints is not None:
                 for constraint in constraints:
                     constrained_fields.update(constraint.eq_fields.keys())
                     constrained_fields.update(constraint.fields.keys())
@@ -577,8 +634,8 @@ class Engine(object):
                 field = item.meta_.fields.get(name)
                 value = getattr(item, name)
                 kwargs = {}
-                if _raise_on_conflict and name not in constrained_fields:
-                    kwargs['expected'] = item.ddb_dump_cached_(name)
+                if raise_on_conflict and name not in constrained_fields:
+                    kwargs = {'eq': item.ddb_dump_cached_(name)}
                 update = ItemUpdate.put(name, item.ddb_dump_field_(name),
                                         **kwargs)
                 updates.append(update)
@@ -615,4 +672,93 @@ class Engine(object):
                 except CheckFailed:
                     pass
         # Refresh item data
-        self.refresh(refresh_models, consistent=consistent)
+        if not no_read:
+            self.refresh(refresh_models, consistent=consistent)
+
+    def update_field(self, item, name, value=NO_ARG, action=ItemUpdate.PUT,
+                     constraints=None):
+        """
+        Update the value of a single field
+
+        Note that this method bypasses field validators and will ignore any
+        special behavior around Composite fields.
+
+        Parameters
+        ----------
+        item : :class:`~flywheel.models.Model`
+            The model to update
+        name : str
+            The name of the field to update
+        value : object, optional
+            The new value for the field. Default will use the value currently
+            on the model.
+        action : str, optional
+            PUT, ADD, or DELETE. (default PUT)
+        constraints : list, optional
+            List of constraints that must pass for the update to complete.
+            Format is the same as query filters (e.g. Model.fieldname > 5)
+
+        """
+        if value is NO_ARG:
+            if action == ItemUpdate.DELETE:
+                value = None
+            elif action == ItemUpdate.ADD:
+                raise ValueError("Must specify a value when using the "
+                                 "actions ADD or DELETE")
+            else:
+                value = getattr(item, name)
+
+        keywords = {}
+        if constraints is not None:
+            for constraint in constraints:
+                keywords.update(constraint.scan_kwargs())
+
+        updates = [
+            ItemUpdate(action, name, item.field_(name).ddb_dump_for_query(value))
+        ]
+
+        ret = self.dynamo.update_item(
+            item.meta_.ddb_tablename(self.namespace), item.pk_dict_,
+            updates, returns=UPDATED_NEW, **keywords)
+        with item.partial_loading_():
+            for key, val in six.iteritems(ret):
+                item.set_ddb_val_(key, val)
+            # If we didn't see the field in the response,
+            # it must have been deleted.
+            if name not in ret:
+                item.set_ddb_val_(name, None)
+
+        item.post_save_fields_(set([name]))
+
+    def exists(self, model, key_or_item, range_key=None, consistent=False):
+        """
+        Check if an item exists in the database
+
+        Parameters
+        ----------
+        model : :class:`dynamodb3.Model`
+            The model class of the item to check
+        key_or_item : dict or :class:`dynamodb3.Model` or object
+            Either the value of the hash key, a model instance, or a dict that
+            contains the primary key.
+        range_key : object, optional
+            Value of the range key (if the previous argument is the hash key)
+        consistent : bool, optional
+            Perform a consistent read from dynamo (default False)
+
+        """
+        if isinstance(key_or_item, Model):
+            pkey = key_or_item.pk_dict_
+        elif isinstance(key_or_item, dict):
+            pkey = model.meta_.pk_dict(None, key_or_item)
+        else:
+            pkey = {
+                model.meta_.hash_key.name: model.meta_.hash_key.ddb_dump(key_or_item)
+            }
+            if range_key is not None:
+                pkey[model.meta_.range_key.name] = range_key
+        attrs = [model.meta_.hash_key.name]
+        ret = self.dynamo.get_item2(
+            model.meta_.ddb_tablename(self.namespace), pkey, attrs,
+            consistent=consistent)
+        return ret.exists
