@@ -1,4 +1,5 @@
 """ Field declarations for models """
+import copy
 import six
 import inspect
 import json
@@ -27,7 +28,7 @@ class Field(object):
     index : str, optional
         If present, create a local secondary index on this field with this as
         the name.
-    data_type : object, optional
+    type : object, optional
         The field data type. You may use int, unicode, set, etc. or you may
         pass in an instance of :class:`~flywheel.fields.types.TypeDefinition`
         (default unicode)
@@ -65,14 +66,15 @@ class Field(object):
     """
 
     def __init__(self, hash_key=False, range_key=False, index=None,
-                 data_type=six.text_type, coerce=False, check=None,
-                 nullable=True, default=NO_ARG):
+                 data_type=NO_ARG, type=six.text_type, coerce=False,
+                 check=None, nullable=True, default=NO_ARG):
         if hash_key and range_key:
             raise ValueError("hash_key and range_key are mutually exclusive!")
         self.name = None
         self.model = None
         self.composite = False
-        self.overflow = False
+        if data_type is NO_ARG:
+            data_type = type
         if isinstance(data_type, TypeDefinition):
             self.data_type = data_type
         elif (inspect.isclass(data_type) and
@@ -101,14 +103,19 @@ class Field(object):
         self._ddb_index_kwargs = None
         if default is NO_ARG:
             if self.is_set:
-                self.default = set()
+                self._default = set()
             else:
-                self.default = None
+                self._default = None
         else:
-            self.default = default
+            self._default = default
         if index:
             self.all_index(index)
 
+    @property
+    def default(self):
+        """ Get a shallow copy of the default value """
+        return copy.copy(self._default)
+
     def get_ddb_index(self):
         """ Construct a dynamo local index object """
         range_key = DynamoKey(self.name, data_type=self.ddb_data_type)
@@ -224,8 +231,6 @@ class Field(object):
         """ Dump a value to format for use in a Dynamo query """
         if value is None:
             return None
-        if self.overflow:
-            return self.ddb_dump_overflow(value)
         value = self.coerce(value, force_coerce=True)
         return self.ddb_dump(value)
 
@@ -233,39 +238,6 @@ class Field(object):
         """ Decode a value retrieved from Dynamo """
         return self.data_type.ddb_load(val)
 
-    @classmethod
-    def ddb_dump_overflow(cls, val):
-        """ Dump an overflow value to its Dynamo format """
-        if val is None:
-            return None
-        elif isinstance(val, six.integer_types) or isinstance(val, float):
-            return val
-        elif isinstance(val, set):
-            return val
-        else:
-            return json.dumps(val)
-
-    @classmethod
-    def ddb_load_overflow(cls, val):
-        """ Decode a value of an overflow field """
-        if (isinstance(val, Decimal) or isinstance(val, float) or
-                isinstance(val, six.integer_types)):
-            if val % 1 == 0:
-                return int(val)
-            return float(val)
-        elif isinstance(val, set):
-            return val
-        else:
-            return json.loads(val)
-
-    @classmethod
-    def is_overflow_mutable(cls, val):
-        """ Check if an overflow field is mutable """
-        val_type = type(val)
-        if val_type in ALL_TYPES:
-            return ALL_TYPES[val_type].mutable
-        return True
-
     @property
     def ddb_data_type(self):
         """ Get the native DynamoDB data type """
@@ -307,6 +279,13 @@ class Field(object):
         else:
             return scope[self.name]
 
+    def get_cached_value(self, obj):
+        """ Get the cached value of a field before any local modifications """
+        if self.name in obj.__cache__:
+            return obj.__cache__[self.name]
+        else:
+            return self.resolve(obj)
+
     def _make_condition(self, filter, other):
         """
         Construct a query condition for a filter on a value
@@ -324,9 +303,6 @@ class Field(object):
             # Don't bother checking allowed filters because this turns into the
             # "exists" filter, which can be done on anything
             pass
-        elif self.overflow:
-            # Don't bother checking validity if this is an overflow field
-            pass
         elif filter not in self.data_type.allowed_filters:
             raise TypeError("Cannot use '%s' filter on '%s' field" %
                             (filter, self.data_type))
@@ -352,7 +328,7 @@ class Field(object):
 
     def _make_contains_condition(self, filter, other):
         """ Construct a query condition for 'contains' or 'ncontains' """
-        if not self.overflow and filter not in self.data_type.allowed_filters:
+        if filter not in self.data_type.allowed_filters:
             raise TypeError("Cannot use '%s' filter on '%s' field" %
                             (filter, self.data_type))
         return Condition.construct(self.name, filter,
@@ -371,9 +347,7 @@ class Field(object):
         Create a query condition that this field must be within a set of values
 
         """
-        if self.overflow:
-            other = set([self.ddb_dump_overflow(val) for val in other])
-        elif 'in' not in self.data_type.allowed_filters:
+        if 'in' not in self.data_type.allowed_filters:
             raise TypeError("Cannot use '%s' filter on '%s' field" %
                             (filter, self.data_type))
         else:
@@ -385,16 +359,10 @@ class Field(object):
         Create a query condition that this field must begin with a string
 
         """
-        if (not self.overflow and
-                'beginswith' not in self.data_type.allowed_filters):
+        if 'beginswith' not in self.data_type.allowed_filters:
             raise TypeError("Cannot use 'beginswith' filter on '%s' field" %
                             self.data_type)
-        if self.overflow:
-            # Since strings are dumped to json in overflow fields, we should
-            # prefix it with a double quote
-            other = '"' + other
-        else:
-            other = self.ddb_dump_for_query(other)
+        other = self.ddb_dump_for_query(other)
         return Condition.construct(self.name, 'beginswith', other)
 
     def between_(self, low, high):
@@ -403,8 +371,7 @@ class Field(object):
         (inclusive)
 
         """
-        if (not self.overflow and
-                'between' not in self.data_type.allowed_filters):
+        if 'between' not in self.data_type.allowed_filters:
             raise TypeError("Cannot use 'between' filter on %s field" %
                             self.data_type)
         low = self.ddb_dump_for_query(low)
@@ -472,3 +439,8 @@ class Composite(Field):
         args = [self.model.meta_.fields[f].resolve(obj, scope) for f in
                 self.subfields]
         return self.coerce(self.merge(*args))
+
+    def get_cached_value(self, obj):
+        args = [self.model.meta_.fields[f].get_cached_value(obj) for f in
+                self.subfields]
+        return self.coerce(self.merge(*args))
