@@ -3,10 +3,11 @@ import six
 import time
 
 import inspect
-from dynamo3 import DynamoKey, LocalIndex, GlobalIndex, Throughput
+from dynamo3 import DynamoKey, Throughput, IndexUpdate
 from collections import defaultdict
 
 from .fields import Field
+from .fields.conditions import FILTER_ONLY
 
 
 class ValidationError(Exception):
@@ -50,6 +51,8 @@ class Ordering(object):
             else:
                 for field in self.range_key.can_resolve(fields):
                     (op, val) = fields[field]
+                    if op in FILTER_ONLY:
+                        continue
                     kwargs['%s__%s' % (field, op)] = val
                     remaining.remove(field)
 
@@ -65,6 +68,25 @@ class Ordering(object):
 
         return kwargs
 
+    def pk_dict(self, obj=None, scope=None, ddb_dump=False):
+        """ Get the dynamo primary key dict for this ordering """
+        # If we can unambiguously tell that a single string defines the primary
+        # key, allow scope to be a single string
+        if (obj is None and isinstance(scope, six.string_types) and
+                self.range_key is None):
+            scope = {self.hash_key.name: scope}
+
+        hk = self.hash_key.resolve(obj, scope)
+        if ddb_dump:
+            hk = self.hash_key.ddb_dump(hk)
+        key_dict = {self.hash_key.name: hk}
+        if self.range_key is not None:
+            rk = self.range_key.resolve(obj, scope)
+            if ddb_dump:
+                rk = self.range_key.ddb_dump(rk)
+            key_dict[self.range_key.name] = rk
+        return key_dict
+
     def __repr__(self):
         if self.range_key is None:
             return "Ordering(%s, None, %s)" % (self.hash_key.name,
@@ -166,8 +188,10 @@ class ModelMetadata(object):
         self._abstract = False
         self.__dict__.update(model.__metadata__)
         # Allow throughput to be specified as read/write in a dict
+        # pylint: disable=E1134
         if isinstance(self.throughput, dict):
             self.throughput = Throughput(**self.throughput)
+        # pylint: enable=E1134
         self.name = self._name
         self.fields = {}
         self.hash_key = None
@@ -308,24 +332,35 @@ class ModelMetadata(object):
         """ Construct the primary key value """
         return self.hash_key.resolve(obj, scope)
 
-    def pk_dict(self, obj=None, scope=None, ddb_dump=False):
-        """ Get the dynamo primary key dict for an item """
-        # If we can unambiguously tell that a single string defines the primary
-        # key, allow scope to be a single string
-        if (obj is None and isinstance(scope, six.string_types) and
-                self.range_key is None):
-            scope = {self.hash_key.name: scope}
-
+    def pk_tuple(self, obj=None, scope=None, ddb_dump=False, ddb_load=False):
+        """ Get a tuple that represents the primary key for an item """
         hk = self.hk(obj, scope)
         if ddb_dump:
             hk = self.hash_key.ddb_dump(hk)
+        elif ddb_load:
+            hk = self.hash_key.ddb_load(hk)
+        if self.range_key is None:
+            return (hk,)
         rk = self.rk(obj, scope)
-        key_dict = {self.hash_key.name: hk}
-        if rk is not None:
-            if ddb_dump:
-                rk = self.range_key.ddb_dump(rk)
-            key_dict[self.range_key.name] = rk
-        return key_dict
+        if ddb_dump:
+            rk = self.range_key.ddb_dump(rk)
+        elif ddb_load:
+            rk = self.range_key.ddb_load(rk)
+        return (hk, rk)
+
+    def pk_dict(self, obj=None, scope=None, ddb_dump=False):
+        """ Get the dynamo primary key dict for an item """
+        return self.index_pk_dict(None, obj, scope, ddb_dump)
+
+    def index_pk_dict(self, index_name, obj=None, scope=None, ddb_dump=False):
+        """ Get the primary key dict for an index (includes the table key) """
+        # Get the 'table' index, which is the hash & range key
+        table_order = self.get_ordering_from_index(None)
+        pk = table_order.pk_dict(obj, scope, ddb_dump)
+        if index_name is not None:
+            ordering = self.get_ordering_from_index(index_name)
+            pk.update(ordering.pk_dict(obj, scope, ddb_dump))
+        return pk
 
     @property
     def abstract(self):
@@ -461,6 +496,83 @@ class ModelMetadata(object):
 
         return tablename
 
+    def update_dynamo_schema(self, connection, test=False,
+                             wait=False, throughput=None, namespace=()):
+        """
+        Updates all Dynamo table global indexes for this model
+
+        Parameters
+        ----------
+        connection : :class:`~dynamo3.DynamoDBConnection`
+        test : bool, optional
+            If True, don't actually create the table (default False)
+        wait : bool, optional
+            If True, block until table has been created (default False)
+        throughput : dict, optional
+            The throughput of the table and global indexes. Has the keys 'read'
+            and 'write'. To specify throughput for global indexes, add the name
+            of the index as a key and another 'read', 'write' dict as the
+            value.
+        namespace : str or tuple, optional
+            The namespace of the table
+
+        Returns
+        -------
+        table : str
+            Table name that altered, or None if nothing altered
+
+        """
+        if self.abstract:
+            return None
+
+        tablename = self.ddb_tablename(namespace)
+
+        global_indexes = []
+
+        for gindex in self.global_indexes:
+            index = gindex.get_ddb_index(self.fields)
+            if throughput is not None and gindex.name in throughput:
+                index.throughput = Throughput(**throughput[gindex.name])
+            global_indexes.append(index)
+
+        if not global_indexes:
+            return None
+
+        table = connection.describe_table(tablename)
+        if not table:
+            return None
+
+        expected_indexes = {}
+        for i in global_indexes:
+            expected_indexes[i.name] = i
+        actual_indexes = {}
+        for i in table.global_indexes:
+            actual_indexes[i.name] = i
+
+        missing_index_names = set(expected_indexes.keys()) - set(actual_indexes.keys())
+        missing_indexes = [expected_indexes[i] for i in missing_index_names]
+
+        updates = [IndexUpdate.create(index) for index in missing_indexes]
+
+        update_indexes_name = set(expected_indexes.keys()) & set(actual_indexes.keys())
+        update_indexes = [expected_indexes[i] for i in update_indexes_name if
+                          actual_indexes[i].throughput != expected_indexes[i].throughput]
+
+        updates.extend([IndexUpdate.update(index.name, index.throughput) for index in update_indexes])
+
+        if not updates:
+            return None
+
+        if not test:
+            connection.update_table(tablename, index_updates=updates)
+            if wait:
+                desc = connection.describe_table(tablename)
+                while desc.status != 'ACTIVE':
+                    time.sleep(1)
+                    desc = connection.describe_table(tablename)
+
+        return tablename
+
     def delete_dynamo_schema(self, connection, tablenames=None, test=False,
                              wait=False, namespace=()):
         """
